;; Generated package description from peg.el  -*- no-byte-compile: t -*-
(define-package "peg" "1.0.1.0.20221212.225433" "Parsing Expression Grammars in Emacs Lisp" '((emacs "25")) :commit "ad531e43d6fbc0ce874f82fcc500f7292a2edd17" :url "https://elpa.gnu.org/packages/peg.html" :authors '(("Helmut Eller" . "eller.helmut@gmail.com")) :maintainer '("Stefan Monnier" . "monnier@iro.umontreal.ca"))
